//////////////////////////////////////////////////////////////////////////////
// Sketch: usb_ghost
// แสดงตัวอย่างการจำลองบอร์ดไมโครคอนโทรลเลอร์ให้เป็นอุปกรณ์เมาส์และคีย์บอร์ดในเวลาเดียวกัน
//
// == การคอมไพล์ ==
// - ต้องการไลบรารี usbdrv และ pt
// - โค้ดนี้ไม่สามารถคอมไพล์ด้วย Arduino IDE ได้ ต้องอาศัย Arduino Makefile ในการคอมไพล์
//
// == การใช้งาน ==
// - กดปุ่มบนบอร์ดพ่วงเพื่อเริ่มต้นส่งคำสั่งเลื่อนเมาส์และการกดคีย์ไปยังโฮสท์
//////////////////////////////////////////////////////////////////////////////
#include <Practicum.h>
#include <pt.h>
#include "keycodes.h"
#include <AcceleroMMA7361.h>

extern "C" {
#include "usbdrv.h"
}

#define REPORT_ID_KEYBOARD 1
#define REPORT_ID_MOUSE    2

#define PIN_RED PIN_PC0
#define PIN_GREEN PIN_PC2
#define PIN_X PIN_PC5
#define PIN_Y PIN_PC1
#define PIN_LEFT PIN_PC4
#define PIN_RIGHT PIN_PC3

// มาโครสำหรับจำลองการหน่วงเวลาใน protothread
#define PT_DELAY(pt,ms,ts) \
  ts = millis(); \
  PT_WAIT_UNTIL(pt, millis()-ts >= (ms));

/////////////////////////////////////////////////////////////////////
// USB report descriptor, สร้างขึ้นจาก HID Descriptor Tool ซึ่ง
// ดาวน์โหลดได้จาก
//    http://www.usb.org/developers/hidpage/dt2_4.zip
//
// *** ไม่แนะนำให้สร้างเองด้วยมือ ***
/////////////////////////////////////////////////////////////////////
PROGMEM const char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = 
{
  0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
  0x09, 0x06,                    // USAGE (Keyboard)
  0xa1, 0x01,                    // COLLECTION (Application)
  0x85, 0x01,                    //   REPORT_ID (1)
  0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
  0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
  0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
  0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
  0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
  0x75, 0x01,                    //   REPORT_SIZE (1)
  0x95, 0x08,                    //   REPORT_COUNT (8)
  0x81, 0x02,                    //   INPUT (Data,Var,Abs)
  0x95, 0x01,                    //   REPORT_COUNT (1)
  0x75, 0x08,                    //   REPORT_SIZE (8)
  0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
  0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
  0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
  0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
  0xc0,                          // END_COLLECTION
  0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
  0x0b, 0x02, 0x00, 0x01, 0x00,  // USAGE (Generic Desktop:Mouse)
  0xa1, 0x01,                    // COLLECTION (Application)
  0x0b, 0x01, 0x00, 0x01, 0x00,  //   USAGE (Generic Desktop:Pointer)
  0xa1, 0x00,                    //   COLLECTION (Physical)
  0x85, 0x02,                    //     REPORT_ID (2)
  0x05, 0x09,                    //     USAGE_PAGE (Button)
  0x1b, 0x01, 0x00, 0x09, 0x00,  //     USAGE_MINIMUM (Button:Button 1)
  0x2b, 0x03, 0x00, 0x09, 0x00,  //     USAGE_MAXIMUM (Button:Button 3)
  0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
  0x25, 0x01,                    //     LOGICAL_MAXIMUM (1)
  0x75, 0x01,                    //     REPORT_SIZE (1)
  0x95, 0x03,                    //     REPORT_COUNT (3)
  0x81, 0x02,                    //     INPUT (Data,Var,Abs)
  0x75, 0x05,                    //     REPORT_SIZE (5)
  0x95, 0x01,                    //     REPORT_COUNT (1)
  0x81, 0x03,                    //     INPUT (Cnst,Var,Abs)
  0x05, 0x01,                    //     USAGE_PAGE (Generic Desktop)
  0x0b, 0x30, 0x00, 0x01, 0x00,  //     USAGE (Generic Desktop:X)
  0x0b, 0x31, 0x00, 0x01, 0x00,  //     USAGE (Generic Desktop:Y)
  0x15, 0x81,                    //     LOGICAL_MINIMUM (-127)
  0x25, 0x7f,                    //     LOGICAL_MAXIMUM (127)
  0x75, 0x08,                    //     REPORT_SIZE (8)
  0x95, 0x02,                    //     REPORT_COUNT (2)
  0x81, 0x06,                    //     INPUT (Data,Var,Rel)
  0xc0,                          //     END_COLLECTION
  0xc0,                          // END_COLLECTION
};

struct ReportKeyboard
{
  /* +----\------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |Byte \ Bit |  7  |  6  |  5  |  4  |  3  |  2  |  1  |  0  |
   * +------\----+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  0        |               Report ID = 1                   |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  1        |           Modifiers (shift,ctrl,etc)          |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  2        |                 Key Code                      |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   */
  uint8_t  report_id;
  uint8_t  modifiers;
  uint8_t  key_code;
};

struct ReportMouse
{
  /* +----\------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |Byte \ Bit |  7  |  6  |  5  |  4  |  3  |  2  |  1  |  0  |
   * +------\----+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  0        |               Report ID = 2                   |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  1        |              Buttons' statuses                |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  2        |                  Delta X                      |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  3        |                  Delta Y                      |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   */
  uint8_t  report_id;
  uint8_t  buttons;
  int8_t   dx;
  int8_t   dy;
};

ReportKeyboard reportKeyboard;
ReportMouse reportMouse;

AcceleroMMA7361 accelero;

int accelero_x;
int accelero_y;
int STABILIZER;

bool isRightPressed;
bool isLeftPressed;

// Protothread states
struct pt keyboard_pt;
struct pt mouse_pt;
struct pt led_pt;
struct pt main_pt;
struct pt left_pt;
struct pt right_pt;

////////////////////////////////////////////////////////////////
// Automatically called by usbpoll() when host makes a request
////////////////////////////////////////////////////////////////
usbMsgLen_t usbFunctionSetup(uchar data[8])
{
  usbRequest_t *rq = (usbRequest_t*)data;

  return 0;  /* return nothing to host for now */
}
//////////////////////////////////////////////////////////////
void sendKey(uint8_t keycode, uint8_t modifiers)
{
  reportKeyboard.key_code = keycode;
  reportKeyboard.modifiers = modifiers;
  usbSetInterrupt((uchar*)&reportKeyboard, sizeof(reportKeyboard));
}

//////////////////////////////////////////////////////////////
void sendMouse(int8_t dx, int8_t dy, uint8_t buttons)
{
  reportMouse.dx = dx;
  reportMouse.dy = dy;
  reportMouse.buttons = buttons;
  usbSetInterrupt((uchar*)&reportMouse, sizeof(reportMouse));
}

//////////////////////////////////////////// //////////////////
PT_THREAD(keyboard_task(struct pt *pt))
{
  static uint32_t ts = 0;

  PT_BEGIN(pt);

  while (1)
  {
    sendKey(KEY_NONE, 0);
    PT_DELAY(pt,1000000,ts);
  }
  
  PT_END(pt);
}

//////////////////////////////////////////////////////////////
PT_THREAD(mouse_task(struct pt *pt))
{
  static uint32_t ts = 0;

  PT_BEGIN(pt);

  while (1)
  {
    accelero_x = accelero.getXAccel();
    accelero_y = accelero.getYAccel();
    STABILIZER = 32;

    if (-STABILIZER < accelero_x && accelero_x < STABILIZER) 
    {
      accelero_x = 0;
    }
    if (-STABILIZER < accelero_y && accelero_y < STABILIZER) 
    {
      accelero_y = 0;
    }

    accelero_y *= 2;

    int factor_x = 1;
    int factor_y = 1;
    if (accelero_x < 0) factor_x = -1;
    if (accelero_y < 0) factor_y = -1;
    int x = factor_x*accelero_x*accelero_x*accelero_x*accelero_x/200000 + accelero_x/32;
    int y = factor_y*accelero_y*accelero_y*accelero_y*accelero_y/200000 + accelero_y/32;

    uint8_t mouseClick = 0;
    if (isLeftPressed) mouseClick = 1;
    else if (isRightPressed) mouseClick = 2;

    sendMouse((int8_t)(x), (int8_t)(y), mouseClick);
    PT_DELAY(pt,30,ts);
  }

  PT_END(pt);
}

//////////////////////////////////////////////////////////////
PT_THREAD(left_task(struct pt *pt))
{
  static uint32_t ts = 0;

  PT_BEGIN(pt);

  while (1)
  {    
    isLeftPressed = false;
    while (digitalRead(PIN_LEFT) == HIGH) 
    {
      isLeftPressed = false;
      PT_DELAY(pt, 20, ts);
    }
    PT_DELAY(pt, 20, ts);
    while (digitalRead(PIN_LEFT) == LOW) 
    {
      isLeftPressed = true;
      PT_DELAY(pt, 20, ts);
    }
    PT_DELAY(pt, 20, ts);
  }

  PT_END(pt);
}
//////////////////////////////////////////////////////////////
PT_THREAD(right_task(struct pt *pt))
{
  static uint32_t ts = 0;

  PT_BEGIN(pt);

  while (1)
  {    
    isRightPressed = false;
    while (digitalRead(PIN_RIGHT) == HIGH) 
    {
      isRightPressed = false;
      PT_DELAY(pt, 20, ts);
    }
    PT_DELAY(pt, 20, ts);
    while (digitalRead(PIN_RIGHT) == LOW) 
    {
      isRightPressed = true;
      PT_DELAY(pt, 20, ts);
    }
    PT_DELAY(pt, 20, ts);
  }

  PT_END(pt);
}
//////////////////////////////////////////////////////////////
PT_THREAD(led_task(struct pt *pt))
{
  static uint32_t ts = 0;

  PT_BEGIN(pt);

  while (1)
  {    
    digitalWrite(PIN_GREEN, LOW);
    PT_DELAY(pt,100,ts);
    digitalWrite(PIN_GREEN, HIGH);
    PT_DELAY(pt,100,ts);
  }

  PT_END(pt);
}
//////////////////////////////////////////////////////////////
PT_THREAD(main_task(struct pt *pt))
{
  PT_BEGIN(pt);

  // Wait until switch is pressed
  PT_WAIT_UNTIL(pt, digitalRead(PIN_RIGHT) == LOW);
  digitalWrite(PIN_GREEN, LOW);
  digitalWrite(PIN_RED, HIGH);
  PT_WAIT_UNTIL(pt, digitalRead(PIN_RIGHT) == HIGH);
  isRightPressed = false;
  isLeftPressed = false;

  while (1)
  {
    led_task(&led_pt);
    right_task(&right_pt);
    left_task(&left_pt);
    // Wait until host is ready for receiving report, then report keyboard and
    // mouse events in a round-robin fashion
    PT_WAIT_UNTIL(pt, usbInterruptIsReady());
    keyboard_task(&keyboard_pt);
    PT_WAIT_UNTIL(pt, usbInterruptIsReady());
    mouse_task(&mouse_pt);
  }

  PT_END(pt);
}

//////////////////////////////////////////////////////////////
void setup()
{
  accelero.begin(13, 12, 11, 10, PIN_X, PIN_Y, 9);
  accelero.setARefVoltage(5); //sets the AREF voltage to 3.3V
  accelero.setSensitivity(LOW); //sets the sensitivity to +/-6G
  accelero.calibrate();

  // Initialize USB subsystem
  usbInit();
  usbDeviceDisconnect();
  delay(300);
  usbDeviceConnect();
  
  // Initialize hardware pins
  pinMode(PIN_RED, OUTPUT);
  pinMode(PIN_GREEN, OUTPUT);
  pinMode(PIN_RIGHT, INPUT_PULLUP);
  pinMode(PIN_LEFT, INPUT_PULLUP);
  pinMode(PIN_X, INPUT);
  pinMode(PIN_Y, INPUT);
  pinMode(PIN_PD3, OUTPUT);

  // Initialize USB reports
  reportKeyboard.report_id = REPORT_ID_KEYBOARD;
  reportKeyboard.modifiers = 0;
  reportKeyboard.key_code  = KEY_NONE;

  reportMouse.report_id = REPORT_ID_MOUSE;
  reportMouse.dx        = 0;
  reportMouse.dy        = 0;
  reportMouse.buttons   = 0;

  // Initialize tasks
  PT_INIT(&keyboard_pt);
  PT_INIT(&mouse_pt);
  PT_INIT(&led_pt);
  PT_INIT(&main_pt);
  PT_INIT(&right_pt);
  PT_INIT(&left_pt);
}

//////////////////////////////////////////////////////////////
void loop()
{
  usbPoll();
  main_task(&main_pt);
}
